// fournit le premier milliard de chiffres décimaux de pi en un peu plus de 35 minutes sur
// macBook Pro M1 2020 16 Go (macOS Ventura 13.4, Xcode 14.3 (14E222b), gmp 6.2.99)
// ligne de commande (avec installation par défaut de gmp) :
// g++ -std=c++17 pi.cpp -opi++ -lgmp -lgmpxx
// ou (avec installation statique de gmp dans <prefix>)
// g++ -std=c++17 pi.cpp -opi++ -I<prefix>/include -L<prefix>/lib -lgmp -lgmpxx
// utilisable aussi sur plateforme GNU/Linux comme ci-dessus
// ou sur plateforme MSYS2/MinGW (Windows) : comme ci-dessus en ajoutant -D_MINGW_ à la ligne de commande

#include <stdio.h> // FILE, fopen(), fclose(), gmpz_out_raw(), gmpz_inp_raw() ...
#include <gmpxx.h> // gmp c++
#include <iostream> // std::cout << ... << std::endl
#include <filesystem> // std::filesystem::current_path() ( -std=c++17 )
#include <iomanip> // std::fixed, std::setprecision()
#include <sstream> // std::stringstream
#include <thread> // std::thread
#include <chrono> // std::chrono::system_clock()

#ifdef _MINGW_
#include <windows.h>
#endif

using namespace std;

unsigned long N; // nombre de décimales de pi calculées (argv[1] de main())
string pi_path; // chemin du répertoire "pi"
void pi(); // fonction définie en fin de fichier appelée par main()

int main (int argc, char* argv[])
{
#ifdef _MINGW_
    SetConsoleOutputCP(65001);
    stringstream ss;
    ss << std::filesystem::current_path();
    pi_path = ss.str().substr(1, ss.str().length() - 2) + "\\\\pi";
#else
    pi_path = string(filesystem::current_path())+"/pi";
#endif
    filesystem::directory_entry entry(pi_path);
    if (!entry.exists()){
        if (!filesystem::create_directory("pi")) {
            cout << "la création du répertoire " << pi_path << " a échoué" << endl;
            exit(EXIT_FAILURE);
        }
    }
    if (argc != 2) {
        std:: cout << "synopsis: " << argv[0] << " < nombre de décimales de pi >\n";
        return 0;
    }
    N = std::stol(argv[1]);
    pi();
}

//---- entrées-sorties

void bigint_to_binfile(const mpz_class bigint, const string& fName)
{
    FILE* f = fopen(fName.c_str(),"w");
#ifdef _MINGW_
    mpz_out_str(f,2,bigint.get_mpz_t());
#else
    mpz_out_raw(f,bigint.get_mpz_t());
#endif
    fclose(f);
}

void bigint_from_binfile(mpz_class &bigint, const string& fName)
{
    FILE* f = fopen(fName.c_str(),"r");
#ifdef _MINGW_
    mpz_inp_str(bigint.get_mpz_t(),f,2);
#else
    mpz_inp_raw (bigint.get_mpz_t(),f);
#endif
    fclose(f);
}

void bigint_to_decfile(const mpz_class bigint, const string& fName)
{
    FILE* f = fopen((string(fName)+".txt").c_str(),"w");
    mpz_out_str(f,10,bigint.get_mpz_t());
    fclose(f);
}

//---- temps courant en secondes

namespace std::chrono {
    double secondes() {
        return duration_cast< microseconds >(system_clock::now().time_since_epoch()).count()/1000000.;
    }
}

double secondes()
{
    return std::chrono::secondes();
}

//---- classe représentant la fonction affine x -> u/w + x * v/w avec u,v,w grands entiers de type mpz_class

class affine {
public:
    mpz_class u,v,w;
    affine(){}
    affine(const string& fName){
        bigint_from_binfile(u,fName+"_a");
        bigint_from_binfile(v,fName+"_b");
        bigint_from_binfile(w,fName+"_c");
    }
    void compose(const affine& af){
        u *=af.w;
        u += v*af.u;
        v *= af.v;
        w *= af.w;
    }
    void reduis(){
        reduis(u.get_mpz_t(),v.get_mpz_t(),w.get_mpz_t());
    }
    void to_binfiles(const string& fName){
        bigint_to_binfile(u,fName+"_a");
        bigint_to_binfile(v,fName+"_b");
        bigint_to_binfile(w,fName+"_c");
    }
private:
    void reduis(mpz_t u, mpz_t v, mpz_t w) // division de u,v,w par leur PGCD
    {
        mpz_t t;
        mpz_init(t);
        mpz_gcd(t,u,v);
        mpz_gcd(t,t,w);
        mpz_divexact(u,u,t);
        mpz_divexact(v,v,t);
        mpz_divexact(w,w,t);
        mpz_clear(t);
    }
};

//---- composée de plusieurs fonctions affines, avec multithreading (paramètre i > 0)

void reduis(affine& af){af.reduis();}

void compose(affine& af, affine (&f)(unsigned long), unsigned long a, unsigned long b, int i = 0)
// af = f(a)of(a+1)o...of(b-1)  (a < b requis)
{
    unsigned long m = (a+b)/2;
    if (b == a+1) {
        af = f(a);
    }
    else {
        affine af2;
        if (i > 1) {
            thread th1(compose,ref(af),ref(f),a,m,i-1);
            thread th2(compose,ref(af2),ref(f),m,b,i-1);
            th1.join();
            th2.join();
        }
        else if (i == 1) {
            compose(af,f,a,m);
            compose(af2,f,m,b);
            thread th1(reduis,ref(af));
            thread th2(reduis,ref(af2));
            th1.join();
            th2.join();
        }
        else {
            compose(af,f,a,m);
            compose(af2,f,m,b);
        }
        af.compose(af2);
    }
}

void etape(int j, affine (&f)(unsigned long), unsigned long a, unsigned long b) // (a < b requis)
{
    cout << "étape " << j << endl;
    
    double sec1 = secondes();
    affine af;
    compose(af,f,a,b,4);
        
    stringstream s;
    s << "pi/etape" << j << "_" << N;
    af.to_binfiles(s.str());

    cout << "durée: " << fixed << setprecision(3) << secondes()-sec1 << " s" << endl;
}

//---- cf. le document pdf joint pour les notations suivantes a(p), b(p), f(p), g(p)

mpz_class a(mpz_class p)
{
    return ((1640*p+3066)*p+1804)*p+330;
}

mpq_class b(unsigned long p)
{
    mpz_class num = 1;
    mpz_class den = 1;
    for (unsigned long j = 4*p+1; j <= 4*p+4; j++)
    {
        num *= j;
        den *= (j<<1)+7;
    }
    den <<= 6;
    den = -den;
    mpq_class frac(num,den);
    frac.canonicalize();
    return frac;
}

affine f(unsigned long p) // x -> a(p)+b(p)*x
{
    affine af;
    af.u = a(p);
    mpq_class frac = b(p);
    af.v = frac.get_num();
    af.w = frac.get_den();
    af.u *= af.w;
    return af;
}

affine g(unsigned long p) // g(p) = f(0)of(1)o...of(p) . (avec p = max(N/3,2))
{
    cout << "composée affine en deux étapes enregistrées:" << endl;
    etape(1,f,0,(p/2)); // (0 < p/2)
    etape(2,f,(p/2),(p+1)); // (p/2 < p+1)

    cout << "composition des deux étapes..." << endl;
    
    double sec1=secondes();
    
    stringstream s1;
    s1 << "pi/etape1_" << N;
    affine af(s1.str());

    stringstream s2;
    s2 << "pi/etape2_" << N;
    affine af2(s2.str());

    af.compose(af2);
    cout << "durée: " << fixed << setprecision(3) << secondes()-sec1 << endl;
    return af;
}

//---- partie entière de (10^N*a)/(b*c) et contrôle de l'approximation de pi

mpz_class Ne10; // initialisée plus loins dans pi() : N^10

mpz_class floor10 (unsigned long c, mpz_class a, mpz_class b)
{
    return (Ne10*a)/(b*c);
}

void controle(mpz_class& res, affine& af, unsigned long p)
{
    bool r = true;
    while(r)
    {
        mpz_class res1,ap;
        ap = a(++p);
        res1 = floor10(105,af.u+ap*af.v,af.w);
        r = (res != res1);
        if (r) {
            cout << "correction" << endl;
            res = res1;
            affine af1=f(p);
            af.compose(af1);
        }
    }
}

//----

void pi()
{
    double sec1 = secondes();
    unsigned long p = N/3;
    if (p < 2) p = 2;
    affine af = g(p);
    
    cout << "valeur approchée de pi par défaut avec " << N << " décimales..." << endl;
    
    mpz_ui_pow_ui(Ne10.get_mpz_t(),10,N);
    mpz_class res = floor10(105,af.u,af.w);
    cout << N << " décimales de pi en " << fixed << setprecision(3) << secondes()-sec1 << endl;
    
    cout << "contrôle..." << endl;
    double sec2 = secondes();
    controle(res,af,p);
    cout << "durée: " << fixed << setprecision(3) << secondes()-sec2 << endl;
        
    cout << "\nenregistrement binaire..." << endl;
    double sec3 = secondes();
    stringstream titre;
    titre << "pi-" << N;
    bigint_to_binfile(res,string("pi/")+titre.str());
    cout << "durée: " << fixed << setprecision(3) << secondes()-sec3 << " s" << endl;
    cout << "durée totale: " << fixed << setprecision(3) << secondes()-sec1 << " s" << endl;
    
    cout << "\nenregistrement décimal..." << endl;
    double sec4 = secondes();
    bigint_to_decfile(res,string("pi/")+titre.str());
    cout << "durée: " << fixed << setprecision(3) << secondes()-sec4 << endl;
    cout << "durée totale: " << fixed << setprecision(3) << secondes()-sec1 << endl;
    cout << "résultat décimal dans le fichier " << titre.str() << ".txt du répertoire " << pi_path << endl;
}
